
from vsg.rules import token_case_with_prefix_suffix

from vsg import token

lTokens = []
lTokens.append(token.package_declaration.identifier)


class rule_010(token_case_with_prefix_suffix):
    '''
    This rule checks the package name has proper case in the package declaration.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       package FIFO_PKG is

    **Fix**

    .. code-block:: vhdl

       package fifo_pkg is
    '''

    def __init__(self):
        token_case_with_prefix_suffix.__init__(self, 'package', '010', lTokens)
        self.groups.append('case::name')




from vsg.rules import single_space_before_token

from vsg.token import case_statement_alternative as token


class rule_005(single_space_before_token):
    '''
    This rule checks for a single space before the **=>** operator.

    **Violation**

    .. code-block:: vhdl

      case data is

        when 3   =>

    **Fix**

    .. code-block:: vhdl

      case data is

        when 3 =>
    '''
    def __init__(self):
        single_space_before_token.__init__(self, 'case', '005', [token.assignment])
        self.solution = 'Reduce spaces before the assignment operator to a single space.'




from vsg import parser

from vsg.rules import single_space_between_tokens

from vsg.token import case_statement_alternative as token


class rule_004(single_space_between_tokens):
    '''
    This rule checks for a single space after the **when** keyword.

    **Violation**

    .. code-block:: vhdl

      case data is

        when   3 =>

    **Fix**

    .. code-block:: vhdl

      case data is

        when 3 =>
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'case', '004', token.when_keyword, parser.todo)
        self.solution = 'Reduce spaces between the *when* keyword and choices to a single space.'




from vsg import parser
from vsg import token


def set_token_indent(dIndentMap, lTokens):
    lTokenKeys, dIndents = process_indent_map(dIndentMap)

    iIndent = 0
    bLibraryFound = False
    bArchitectureFound = False
    dVars = {}
    dVars['insideConcurrentSignalAssignment'] = False

    for oToken in lTokens:

        if isinstance(oToken, parser.whitespace):
            continue

        if isinstance(oToken, parser.blank_line):
            bLibraryFound = False
            continue

        if isinstance(oToken, parser.carriage_return):
            continue

        sUniqueId = oToken.get_unique_id(sJoin=':')

        iTokenIndent = None

        if sUniqueId in lTokenKeys:
            token_key = dIndents[sUniqueId]['token']
            after_key = dIndents[sUniqueId]['after']
            iTokenIndent = update_indent_var(iIndent, token_key)
            oToken.set_indent(iTokenIndent)
            iIndent = update_indent_var(iIndent, after_key)

        if isinstance(oToken, token.context_declaration.end_keyword):
            bLibraryFound = False
            continue

        if isinstance(oToken, token.library_clause.keyword):
            bLibraryFound = True
            continue

        if isinstance(oToken, token.use_clause.keyword):
            if not bArchitectureFound:
                oToken.set_indent(iIndent + 1)
            else:
                oToken.set_indent(iIndent)
            continue

        if isinstance(oToken, token.context_reference.keyword):
            if bLibraryFound:
                oToken.set_indent(iIndent + 1)
            else:
                oToken.set_indent(iIndent)
            continue

        if isinstance(oToken, token.architecture_body.architecture_keyword):
            bLibraryFound = False
            bArchitectureFound = True
            continue

        if isinstance(oToken, token.architecture_body.semicolon):
            bArchitectureFound = False
            continue

        if isinstance(oToken, token.entity_declaration.entity_keyword):
            bLibraryFound = False
            continue

        if isinstance(oToken, token.package_body.package_keyword):
            bLibraryFound = False
            continue

        if isinstance(oToken, token.package_declaration.package_keyword):
            bLibraryFound = False
            continue

        ### Comments
        if isinstance(oToken, parser.comment):
            if bLibraryFound:
                oToken.set_indent(iIndent + 1)
            elif oToken.is_block_comment:
                if oToken.block_comment_indent == 0:
                    oToken.set_indent(0)
                else:
                    oToken.set_indent(iIndent)
            else:
                oToken.set_indent(iIndent)
            continue

        ### Concurrent signal assignment
        if isinstance(oToken, token.concurrent_signal_assignment_statement.label_name):
            dVars['insideConcurrentSignalAssignment'] = True
            continue

        if isinstance(oToken, token.concurrent_signal_assignment_statement.postponed_keyword):
            dVars['insideConcurrentSignalAssignment'] = True
            continue

        if isinstance(oToken, token.concurrent_simple_signal_assignment.target):
            dVars['insideConcurrentSignalAssignment'] = True
            continue

        if isinstance(oToken, token.concurrent_conditional_signal_assignment.target):
            dVars['insideConcurrentSignalAssignment'] = True
            continue

        if isinstance(oToken, token.concurrent_selected_signal_assignment.with_keyword):
            dVars['insideConcurrentSignalAssignment'] = True
            continue

        if isinstance(oToken, token.concurrent_simple_signal_assignment.semicolon):
            dVars['insideConcurrentSignalAssignment'] = False
            continue

        if isinstance(oToken, token.concurrent_conditional_signal_assignment.semicolon):
            dVars['insideConcurrentSignalAssignment'] = False
            continue

        if isinstance(oToken, token.concurrent_selected_signal_assignment.semicolon):
            dVars['insideConcurrentSignalAssignment'] = False
            continue

        oToken.set_indent(iTokenIndent)


def update_indent_var(iIndent, update):
    if str(update).startswith('+'):
        return iIndent + int(update)
    if str(update).startswith('-'):
        return iIndent + int(update)
    if update == 'current':
        return iIndent
    return update


def process_indent_map(dIndentMap):
    dReturn = {}
    lTopKeys = list(dIndentMap['indent']['tokens'].keys())
    for sTopKey in lTopKeys:
        lSubKeys = list(dIndentMap['indent']['tokens'][sTopKey].keys())
        for sSubKey in lSubKeys:
            sUniqueId = sTopKey + ':' + sSubKey
            dReturn[sUniqueId] = dIndentMap['indent']['tokens'][sTopKey][sSubKey]

    lReturn = list(dReturn.keys())

    return lReturn, dReturn




from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure


class insert_token_right_of_token_if_it_does_not_exist_before_token(structure.Rule):
    '''
    Checks for the existence of a token and will insert it if it does not exist.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    insert_token : token object
       token to insert if it does not exist.

    anchor_token : token object type
       token to check if insert_token exists to the right of

    end_token : token object type
       token that bounds the search for the insert_token
    '''

    def __init__(self, name, identifier, insert_token, anchor_token, end_token):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.insert_token = insert_token
        self.anchor_token = anchor_token
        self.end_token = end_token
        self.action = 'add'
        self.configuration.append('action')

    def _get_tokens_of_interest(self, oFile):
        if self.action == 'add':
            return oFile.get_tokens_bounded_by(self.anchor_token, self.end_token)
        else:
            return oFile.get_token_and_n_tokens_before_it([self.insert_token], 1)

    def _analyze(self, lToi):
        if self.action == 'remove':
            for oToi in lToi:
                sSolution = self.action.capitalize() + ' ' + self.solution
                self.add_violation(violation.New(oToi.get_line_number(), oToi, sSolution))
            return

        for oToi in lToi:
            lTokens = oToi.get_tokens()
            bFound = False
            for oToken in lTokens:
                if isinstance(oToken, type(self.insert_token)):
                    bFound = True
                    break
            if not bFound:
                sSolution = self.action.capitalize() + ' ' + self.solution
                self.add_violation(violation.New(oToi.get_line_number(), oToi, sSolution))

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        if self.action == 'remove':
            rules_utils.remove_optional_item(oViolation, self.insert_token)
        else:
            if isinstance(lTokens[1], parser.whitespace) and isinstance(lTokens[2], parser.semicolon):
                rules_utils.insert_token(lTokens, 2, self.insert_token)
            else:
                rules_utils.insert_token(lTokens, 1, self.insert_token)
                rules_utils.insert_whitespace(lTokens, 1)
            oViolation.set_tokens(lTokens)





from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment


class align_consecutive_lines_starting_with_a_comment_above_line_starting_with_token(alignment.Rule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    token : token object
       reference token to align comments with
    '''

    def __init__(self, name, identifier, token, bIncrement=False):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.phase = 4
        self.subphase = 2
        self.token = token
        self.bIncrement = bIncrement

    def analyze(self, oFile):
        self._print_debug_message('Analyzing rule: ' + self.unique_id)
        lToi = oFile.get_consecutive_lines_starting_with_token_and_stopping_when_token_starting_line_is_found(parser.comment, self.token)
        lUpdate = []
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iLine = oToi.get_line_number()

            iTargetIndent = lTokens[-1].get_indent()
            iWhitespaceLength = iTargetIndent * self.indentSize

            lUpdate = []

            for iIndex in range(0, len(lTokens)):
               oToken = lTokens[iIndex]

               if isinstance(oToken, self.token):
                   break

               if iIndex == 0:
                   if isinstance(lTokens[0], parser.whitespace):
                       if len(lTokens[0].get_value()) != iWhitespaceLength:
                           oLineTokens = oFile.get_tokens_from_line(iLine)
                           oViolation = violation.New(iLine, oLineTokens, self.solution)
                           dAction = {}
                           dAction['type'] = 'adjust'
                           dAction['adjust'] = iWhitespaceLength
                           oViolation.set_action(dAction)
                           self.add_violation(oViolation)
                       self._adjust_token_indent(lTokens[1], iTargetIndent)
                   else:
                       oLineTokens = oFile.get_tokens_from_line(iLine)
                       oViolation = violation.New(iLine, oLineTokens, self.solution)
                       dAction = {}
                       dAction['type'] = 'insert'
                       dAction['adjust'] = iWhitespaceLength
                       oViolation.set_action(dAction)
                       self.add_violation(oViolation)
                       self._adjust_token_indent(lTokens[0], iTargetIndent)

               if isinstance(oToken, parser.carriage_return):
                   iLine += 1
                   if isinstance(lTokens[iIndex + 1], parser.whitespace):
                       if len(lTokens[iIndex + 1].get_value()) != iWhitespaceLength:
                           oLineTokens = oFile.get_tokens_from_line(iLine)
                           oViolation = violation.New(iLine, oLineTokens, self.solution)
                           dAction = {}
                           dAction['type'] = 'adjust'
                           dAction['adjust'] = iWhitespaceLength
                           oViolation.set_action(dAction)
                           self.add_violation(oViolation)
                       self._adjust_token_indent(lTokens[iIndex + 2], iTargetIndent)
                   else:
                       oLineTokens = oFile.get_tokens_from_line(iLine)
                       oViolation = violation.New(iLine, oLineTokens, self.solution)
                       dAction = {}
                       dAction['type'] = 'insert'
                       dAction['adjust'] = iWhitespaceLength
                       oViolation.set_action(dAction)
                       self.add_violation(oViolation)
                       self._adjust_token_indent(lTokens[iIndex + 1], iTargetIndent)

            lUpdate.append(violation.New(0, oToi, ''))

        oFile.update(lUpdate)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        if dAction['type'] == 'adjust':
            lTokens[0].set_value(' ' * dAction['adjust'])
        elif dAction['type'] == 'insert':
            rules_utils.insert_whitespace(lTokens, 0, dAction['adjust'])
        oViolation.set_tokens(lTokens)

    def _adjust_token_indent(self, oToken, iIndent):
        oToken.set_indent(iIndent)




from vsg import token

from vsg.rules import single_space_between_token_pairs

lTokens = []
lTokens.append([token.process_statement.process_keyword, token.process_statement.is_keyword])
lTokens.append([token.process_statement.close_parenthesis, token.process_statement.is_keyword])


class rule_014(single_space_between_token_pairs):
    '''
    This rule checks for a single space before the **is** keyword.

    **Violation**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        )     is
       begin

    **Fix**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is
       begin
    '''
    def __init__(self):
        single_space_between_token_pairs.__init__(self, 'process', '014', lTokens)
        self.solution = 'Ensure a single space before the *is* keyword.'




from vsg.rules import single_space_between_tokens

from vsg.token import context_declaration as token


class rule_002(single_space_between_tokens):
    '''
    This rule checks for a single space between the **context** keyword and the context identifier.

    **Violation**

    .. code-block:: vhdl

       context   c1 is

    **Fix**

    .. code-block:: vhdl

       context c1 is
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'context', '002', token.context_keyword, token.identifier)
        self.solution = 'Reduce spaces between *context* keyword and identifier.'




from vsg import parser
from vsg import token

from vsg.vhdlFile import utils


def remove_optional_item(oViolation, oInsertToken):
    lTokens = oViolation.get_tokens()
    if isinstance(lTokens[0], parser.whitespace):
        oViolation.set_tokens([])
    else:
        oViolation.set_tokens([lTokens[0]])


def is_single_line(oToi):
    lTokens = oToi.get_tokens()
    for oToken in lTokens:
        if isinstance(oToken, parser.carriage_return):
            return False
    return True


def number_of_carriage_returns(lTokens):
    iReturn = 0
    for oToken in lTokens:
        if isinstance(oToken, parser.carriage_return):
            iReturn += 1
    return iReturn


def number_of_tokens_from_token_list_in_token_list(lTokens, lTokenList):
    iReturn = 0
    for oToken in lTokens:
        iReturn += number_of_tokens_in_token_list(oToken, lTokenList)
    return iReturn


def number_of_tokens_in_token_list(oToken, lTokens):
    iReturn = 0
    for oTokenItem in lTokens:
        if oToken == type(oTokenItem):
            iReturn += 1
    return iReturn


def does_line_start_with_comment(lTokens):
    if isinstance(lTokens[0], parser.comment):
        return True
    if isinstance(lTokens[0], parser.whitespace) and isinstance(lTokens[1], parser.comment):
        return True
    return False


def get_toi_parameters(oToi):
    return oToi.get_line_number(), oToi.get_tokens()


def insert_token(lTokens, index, oToken):
    try:
        oToken = update_code_tags(lTokens[index], oToken)
    except IndexError:
        oToken = update_code_tags(lTokens[0], oToken)

    lTokens.insert(index, oToken)


def append_token(lTokens, oToken):
    oToken = update_code_tags(lTokens[-1], oToken)
    lTokens.append(oToken)


def update_code_tags(oToken1, oToken2):
    oToken2.code_tags = oToken1.code_tags
    return oToken2


def insert_whitespace(lTokens, index, num=1):
    insert_token(lTokens, index, parser.whitespace(' '*num))


def insert_carriage_return(lTokens, index):
    insert_token(lTokens, index, parser.carriage_return())


def insert_blank_line(lTokens, index):
    insert_token(lTokens, index, parser.blank_line())


def append_whitespace(lTokens, num=1):
    append_token(lTokens, parser.whitespace(' '*num))


def append_carriage_return(lTokens):
    append_token(lTokens, parser.carriage_return())


def get_index_of_token_in_list(oToken, lTokens):
    for iToken, token in enumerate(lTokens):
        if isinstance(token, oToken):
            return iToken
    return None


def get_number_of_carriage_returns_before_token(oStopToken, lTokens):
    iReturn = 0
    for oToken in lTokens:
        if isinstance(oToken, parser.carriage_return):
            iReturn += 1
        if isinstance(oToken, oStopToken):
            break
    return iReturn


def whitespace_before_token_index(lTokens, iIndex):
    if isinstance(lTokens[iIndex - 1], parser.whitespace):
        return True
    return False


def remove_token_sequence_from_token_list(lRemoveTokens, lTokens):
    lSliceIndexes = find_slice_indexes_of_token_sequence_in_token_list(lRemoveTokens, lTokens)
    return remove_slices_from_token_list(lTokens, lSliceIndexes)


def find_slice_indexes_of_token_sequence_in_token_list(lRemoveTokens, lTokens):
    iTokenListLength = len(lTokens)
    iLength = len(lRemoveTokens)
    lReturn = []
    for iIndex in range(0, iTokenListLength - 1 - iLength):
        if is_token_sequence_at_index_in_token_list(lRemoveTokens, iIndex, lTokens):
            lReturn.append([iIndex, iIndex + iLength])
    return lReturn


def is_token_sequence_at_index_in_token_list(lRemoveTokens, iIndex, lTokens):
    iLength = len(lRemoveTokens)
    for iToken in range(0, iLength):
        if not isinstance(lTokens[iIndex + iToken], lRemoveTokens[iToken]):
            return False
    return True


def remove_slices_from_token_list(lTokens, lSlices):
    lReturn = lTokens
    lSlices.reverse()
    for lSlice in lSlices:
        lReturn = remove_slice_from_token_list(lReturn, lSlice)
    return lReturn


def remove_slice_from_token_list(lTokens, lSlice):
    iStart = lSlice[0]
    iEnd = lSlice[1]
    return lTokens[0:iStart] + lTokens[iEnd::]


def token_is_carriage_return(oToken):
    if isinstance(oToken, parser.carriage_return):
        return True
    return False


def token_is_comment(oToken):
    if isinstance(oToken, parser.comment):
        return True
    return False


def token_is_blank_line(oToken):
    if isinstance(oToken, parser.blank_line):
        return True
    return False


def token_is_open_paren(oToken):
    if isinstance(oToken, parser.open_parenthesis):
        return True
    return False


def token_is_close_paren(oToken):
    if isinstance(oToken, parser.close_parenthesis):
        return True
    return False


def token_is_whitespace(oToken):
    if isinstance(oToken, parser.whitespace):
        return True
    return False


def token_list_begins_with_close_paren(lTokens):
    if isinstance(lTokens[0], parser.whitespace) and isinstance(lTokens[1], parser.close_parenthesis):
        return True
    if isinstance(lTokens[0], parser.close_parenthesis):
        return True
    return False


def token_list_starts_with_paren(lTokens, iIndex=0):
    iToken = utils.find_next_non_whitespace_token(iIndex, lTokens)
    if isinstance(lTokens[iToken], parser.open_parenthesis):
        return True
    return False


def token_at_the_beginning_of_a_line(oToken, lTokens):
    if isinstance(lTokens[0], parser.whitespace) and isinstance(lTokens[1], oToken):
        return True
    if isinstance(lTokens[0], oToken):
        return True
    return False


def token_list_is_the_beginning_of_a_line(lTokens):
    if isinstance(lTokens[0], parser.carriage_return) and isinstance(lTokens[1], parser.whitespace):
        return True
    if isinstance(lTokens[1], parser.carriage_return):
        return True
    return False


def left_most_token_is_at_the_end_of_a_line(lTokens):
    for oToken in lTokens[1:3]:
        if token_is_carriage_return(oToken) or token_is_comment(oToken):
            return True
    return False


def whitespace_is_larger_than_a_single_character(lTokens):
    if lTokens[1].get_value() != ' ':
        return True
    return False


def remove_toi_if_token_is_at_the_end_of_the_line(lToi):
    lReturn = []
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if left_most_token_is_at_the_end_of_a_line(lTokens):
            continue
        lReturn.append(oToi)
    return lReturn


def extract_identifiers_with_mode_of_input(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.in_keyword)


def extract_identifiers_with_mode_of_out(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.out_keyword)


def extract_identifiers_with_mode_of_inout(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.inout_keyword)


def extract_identifiers_with_mode_of_buffer(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.buffer_keyword)


def extract_identifiers_with_mode_of_linkage(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.linkage_keyword)


def extract_identifiers_with_mode(lToi, oTokenType):
    lReturn = []
    for oToi in lToi:
        if oToi.token_type_exists(oTokenType):
            lReturn.append(oToi.extract_tokens(0, 0))
    return lReturn




from vsg import parser
from vsg import token

from vsg.rules import single_space_between_token_pairs

lTokens = []
lTokens.append([parser.close_parenthesis, token.if_statement.then_keyword])


class rule_004(single_space_between_token_pairs):
    '''
    This rule checks for a single space between the ) and the **then** keyword.

    **Violation**

    .. code-block:: vhdl

      if (a = '1')then

      if (a = '1')    then

    **Fix**

    .. code-block:: vhdl

      if (a = '1') then

      if (a = '1') then
    '''
    def __init__(self):
        single_space_between_token_pairs.__init__(self, 'if', '004', lTokens)
        self.solution = 'Ensure only a single space exists between the ) and *then* keyword.'




from vsg import parser

from vsg.rules import n_spaces_after_tokens

lTokens = []
lTokens.append(parser.comma)


class rule_007(n_spaces_after_tokens):
    '''
    This rule checks for spaces after a comma.

    **Violation**

    .. code-block:: vhdl

       proc : process (wr_en,rd_en,overflow) is

    **Fix**

    .. code-block:: vhdl

       proc : process (wr_en, rd_en, overflow) is
    '''

    def __init__(self):
        n_spaces_after_tokens.__init__(self, 'whitespace', '007', 1, lTokens, bNIsMinimum=True)
        self.solution = 'Add a space after comma.'




from vsg import parser
from vsg import violation

from vsg.rule_group.blank_line import Rule
from vsg.rules import utils


class rule_200(Rule):
    '''
    This rule removes consecutive blank lines.

    **Violation**

    .. code-block:: vhdl

       a <= b;


       c <= d;

    **Fix**

    .. code-block:: vhdl

       a <= b;

       c <= d;
    '''

    def __init__(self):
        Rule.__init__(self, name='whitespace', identifier='200')
        self.blank_lines_allowed = 1
        self.configuration.append('blank_lines_allowed')

    def _get_tokens_of_interest(self, oFile):
        lToi = oFile.get_consecutive_lines_starting_with_token(parser.blank_line)
        lReturn = []
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if len(lTokens) > ((self.blank_lines_allowed * 2) - 1):
                lReturn.append(oToi)
        return lReturn

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iBlankLines = utils.number_of_tokens_in_token_list(parser.blank_line, lTokens)

            if iBlankLines > self.blank_lines_allowed:
                sSolution = f'Remove {iBlankLines - self.blank_lines_allowed}'
                oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
                dAction = {}
                dAction['remove_index'] = len(lTokens) - (iBlankLines - self.blank_lines_allowed) * 2
                oViolation.set_action(dAction)
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        lNewTokens = lTokens[0:dAction['remove_index']]
        oViolation.set_tokens(lNewTokens)




from vsg.rules import move_token_next_to_another_token

from vsg.token import architecture_body as token


class rule_006(move_token_next_to_another_token):
    '''
    This rule checks the **is** keyword is on the same line as the **architecture** keyword.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo
         is

       architecture rtl of fifo

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is

       architecture rtl of fifo is
    '''

    def __init__(self):
        move_token_next_to_another_token.__init__(self, 'architecture', '006', token.entity_name, token.is_keyword)
        self.solution = 'Ensure *is* keyword is on the same line as the entity name.'




from vsg import parser

from vsg.rules import single_space_between_tokens

from vsg.token import case_statement as token


class rule_002(single_space_between_tokens):
    '''
    This rule checks for a single space after the **case** keyword.

    **Violation**

    .. code-block:: vhdl

       case    data is


    **Fix**

    .. code-block:: vhdl

       case data is
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'case', '002', token.case_keyword, parser.todo)
        self.solution = 'Reduce spaces between *case* keyword and expression.'




from vsg.rules import multiline_alignment_between_tokens

from vsg import token

lTokenPairs = []
lTokenPairs.append([token.simple_variable_assignment.assignment, token.simple_variable_assignment.semicolon])
lTokenPairs.append([token.conditional_variable_assignment.assignment, token.conditional_variable_assignment.semicolon])


class rule_004(multiline_alignment_between_tokens):
    '''
    This rule checks the alignment of multiline variable assignments.

    **Violation**

    .. code-block:: vhdl

         counter := 1 + 4 + 10 + 25 +
              30 + 35;

    **Fix**

    .. code-block:: vhdl

         counter := 1 + 4 + 10 + 25 +
                    30 + 35;
    '''

    def __init__(self):
        multiline_alignment_between_tokens.__init__(self, 'variable_assignment', '004', lTokenPairs)
        self.phase = 5
        self.subphase = 2




from vsg.rules import single_space_after_token

from vsg import token

lTokens = []
lTokens.append(token.constant_declaration.colon)


class rule_005(single_space_after_token):
    '''
    This rule checks for a single space after the colon.

    **Violation**

    .. code-block:: vhdl

       constant size  :integer := 1;
       constant width :     integer := 32;

    **Fix**

    .. code-block:: vhdl

       constant size  : integer := 1;
       constant width : integer := 32;
    '''

    def __init__(self):
        single_space_after_token.__init__(self, 'constant', '005', lTokens)




from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment
from vsg.vhdlFile import utils


class align_left_token_with_right_token_if_right_token_starts_a_line(alignment.Rule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : token object list
       List of tokens to align

    left_token : token object
       The first token that defines the region

    right_token : token object
       The second token that defines the region
    '''

    def __init__(self, name, identifier, left_token, right_token):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.left_token = left_token
        self.right_token = right_token

    def analyze(self, oFile):

        lToi = oFile.get_tokens_bounded_by(self.left_token, self.right_token)
        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)

            if not utils.does_token_type_exist_in_list_of_tokens(parser.carriage_return, lTokens):
                continue

            if not utils.does_token_start_line(len(lTokens) - 1, lTokens):
                continue

            iStartIndex = oToi.get_start_index()
            iEndIndex = oToi.get_end_index()

            iRightColumn = oFile.get_column_of_token_index(iStartIndex)
            iLeftColumn = oFile.get_column_of_token_index(iEndIndex)

            if iRightColumn + 1 != iLeftColumn:
                iLineNumber = iLine + utils.count_token_types_in_list_of_tokens(parser.carriage_return, lTokens)
                sSolution = 'Move ' + lTokens[-1].get_value() + ' to column ' + str(iRightColumn)
                dAction = {}
                if iLeftColumn == 1:
                    dAction['action'] = 'insert'
                else:
                    dAction['action'] = 'adjust'
                dAction['column'] = iRightColumn
                oViolation = violation.New(iLineNumber, oToi, sSolution)
                oViolation.set_action(dAction)
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        if dAction['action'] == 'insert':
            rules_utils.insert_whitespace(lTokens, len(lTokens) - 1, dAction['column'])
        else:
            lTokens[-2].set_value(' '*dAction['column'])
        oViolation.set_tokens(lTokens)




from vsg.rules import single_space_between_tokens

from vsg.token import context_declaration as token


class rule_018(single_space_between_tokens):
    '''
    This rule checks for a single space between the **end** keyword and the **context** keyword.

    **Violation**

    .. code-block:: vhdl

       end;

       end   context;

    **Fix**

    .. code-block:: vhdl

       end;

       end context;
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'context', '018', token.end_keyword, token.end_context_keyword)
        self.solution = 'Reduce spaces between *end* keyword and *context* keyword to a single space.'




from vsg import token

from vsg.rules import single_space_after_token

lTokens = []
lTokens.append(token.attribute_specification.attribute_keyword)
lTokens.append(token.attribute_specification.attribute_designator)
lTokens.append(token.attribute_specification.of_keyword)
lTokens.append(token.attribute_specification.is_keyword)


class rule_100(single_space_after_token):
    '''
    This rule checks for a single space after the following attribute_specification elements:  **attribute** keyword, *attribute_designator*, **of** keyword and **is** keyword.

    **Violation**

    .. code-block:: vhdl

       attribute   coordinate   of   comp_1:component   is   (0.0, 17.5);

       attribute   coordinate   of   comp_1:component   is(0.0, 17.5);

    **Fix**

    .. code-block:: vhdl

       attribute coordinate of comp_1:component   is (0.0, 17.5);

       attribute coordinate of comp_1:component   is (0.0, 17.5);
    '''
    def __init__(self):
        single_space_after_token.__init__(self, 'attribute_specification', '100', lTokens)




from vsg.rules import multiline_alignment_between_tokens as Rule

from vsg import parser
from vsg import token
from vsg.vhdlFile import utils
from vsg.rules import utils as rules_utils

lTokenPairs = []
lTokenPairs.append([token.constant_declaration.assignment_operator, token.constant_declaration.semicolon])


class rule_014(Rule):
    '''
    This rule checks the indent of multiline constants that do not contain arrays.

    |configuring_multiline_indent_rules_link|

    **Violation**

    .. code-block:: vhdl

       constant width : integer := a + b +
         c + d;

    **Fix**

    .. code-block:: vhdl

       constant width : integer := a + b +
                                   c + d;
    '''

    def __init__(self):
        Rule.__init__(self, 'constant', '014', lTokenPairs)
        self.subphase = 3
        self.phase = 5

    def _get_tokens_of_interest(self, oFile):
        for lTokenPair in self.lTokenPairs:
            lToi = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1])
            lToi = remove_arrays(lToi)
        return lToi


def remove_arrays(lToi):
    lReturn = []
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if not rules_utils.token_list_starts_with_paren(lTokens, 1):
            lReturn.append(oToi)
    return lReturn




from vsg.rules import single_space_between_tokens

from vsg.token import entity_declaration as token


class rule_007(single_space_between_tokens):
    '''
    This rule checks for a single space before the **is** keyword.

    **Violation**

    .. code-block:: vhdl

       entity fifo    is

    **Fix**

    .. code-block:: vhdl

       entity fifo is
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'entity', '007', token.identifier, token.is_keyword)
        self.solution = 'Reduce spaces between identifier and *is* keyword to a single space.'




from vsg.rules import single_space_between_tokens

from vsg.token import entity_declaration as token


class rule_011(single_space_between_tokens):
    '''
    This rule checks for a single space after the **end** keyword.

    **Violation**

    .. code-block:: vhdl

       end    entity fifo;

    **Fix**

    .. code-block:: vhdl

       end entity fifo;
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'entity', '011', token.end_keyword, token.end_entity_keyword)
        self.solution = 'Reduce spaces between *end* keyword and *entity* keyword to a single space.'




from vsg.rules import single_space_between_tokens

from vsg.token import architecture_body as token


class rule_012(single_space_between_tokens):
    '''
    This rule checks for a single space between **end** and **architecture** keywords.

    **Violation**

    .. code-block:: vhdl

       end    architecture architecture_name;

    **Fix**

    .. code-block:: vhdl

       end architecture architecture_name;
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'architecture', '012', token.end_keyword, token.end_architecture_keyword)
        self.solution = 'Reduce spaces between *end* and *architecture* keywords to a single space.'




from vsg.rules import single_space_before_token as Rule


class single_space_before_token_if_on_same_line_as_token(Rule):
    '''
    Checks for a single space before a token.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    oFirstToken : token object
       Token to check for a single space before.

    oSecondToken : token object
       Token to use as an exclusion.
    '''

    def __init__(self, name, identifier, oFirstToken, oExceptToken):
        Rule.__init__(self, name=name, identifier=identifier, lTokens=[])
        self.oFirstToken = oFirstToken
        self.oExceptToken = oExceptToken

    def _get_tokens_of_interest(self, oFile):
        lExcept = oFile.get_line_which_includes_tokens([self.oExceptToken])
        lFilter = []
        for oExcept in lExcept:
            lFilter.append(oExcept.get_line_number())
        lReturn = []
        lToi = oFile.get_token_and_n_tokens_before_it([self.oFirstToken], 2)
        for oToi in lToi:
            if oToi.get_line_number() in lFilter:
                lReturn.append(oToi)
        return lReturn





from vsg import parser
from vsg import violation

from vsg.rule_group import whitespace
from vsg.rules import utils as rules_utils


class whitespace_before_token(whitespace.Rule):
    '''
    Checks for a at least a single space before a token.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    token : token object
       reference token check for a whitespace before
    '''

    def __init__(self, name, identifier, lTokens):
        whitespace.Rule.__init__(self, name=name, identifier=identifier)
        self.lTokens = lTokens

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_token_and_n_tokens_before_it(self.lTokens, 1)

    def _analyze(self, lToi):

        for oToi in lToi:
            lTokens = oToi.get_tokens()

            if isinstance(lTokens[0], parser.whitespace):
                continue

            if isinstance(lTokens[0], parser.carriage_return):
                continue

            oViolation = violation.New(oToi.get_line_number(), oToi, self.solution)
            self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        rules_utils.insert_whitespace(lTokens, 1)
        oViolation.set_tokens(lTokens)




from vsg.rules import split_line_at_token

from vsg import token

lTokens = []
lTokens.append(token.entity_declaration.end_keyword)


class rule_021(split_line_at_token):
    '''
    This rule checks the **end** keyword is on it's own line.

    **Violation**

    .. code-block:: vhdl

       port (
          ...
       ); end entity;

    **Fix**

    .. code-block:: vhdl

       port (
          ...
       );
       end entity;
    '''

    def __init__(self):
        split_line_at_token.__init__(self, 'entity', '021', lTokens)
        self.solution = 'Move the **end** keyword to the next line.'




from vsg.rules import single_space_between_tokens

from vsg.token import loop_statement as token


class rule_100(single_space_between_tokens):
    '''
    This rule checks that a single space exists between the **end** and **loop** keywords

    **Violation**

    .. code-block:: vhdl

         end loop;
         end    loop;

    **Fix**

    .. code-block:: vhdl

         end loop;
         end loop;
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'loop_statement', '100', token.end_keyword, token.end_loop_keyword)
        self.solution = 'Ensure a single space is between *end* and *loop*.'




from vsg.rules import single_space_between_tokens

from vsg.token import case_statement as token


class rule_006(single_space_between_tokens):
    '''
    This rule checks for a single space between the **end** and **case** keywords.

    **Violation**

    .. code-block:: vhdl

      case data is

      end    case;

    **Fix**

    .. code-block:: vhdl

      case data is

      end case;
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'case', '006', token.end_keyword, token.end_case_keyword)
        self.solution = 'Reduce spaces between the *end* and *case* keywords to one space.'




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.entity_declaration.end_entity_keyword)


class rule_014(token_case):
    '''
    This rule checks the **entity** keyword has proper case in the closing of the entity declaration.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       end ENTITY fifo;

    **Fix**

    .. code-block:: vhdl

       end entity fifo;
    '''

    def __init__(self):
        token_case.__init__(self, 'entity', '014', lTokens)
        self.groups.append('case::keyword')




from .rule_001 import rule_001
from .rule_002 import rule_002

from .rule_100 import rule_100

from .rule_300 import rule_300
from .rule_301 import rule_301
from .rule_302 import rule_302

from .rule_500 import rule_500
from .rule_501 import rule_501
#
#from .rule_600 import rule_600
#from .rule_601 import rule_601




from vsg.rules import single_space_between_tokens

from vsg.token import architecture_body as token


class rule_022(single_space_between_tokens):
    '''
    This rule checks for a single space before the entity name in the end architecture declaration.

    **Violation**

    .. code-block:: vhdl

       end architecture    fifo;

    **Fix**

    .. code-block:: vhdl

       end architecture fifo;
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'architecture', '022', token.end_architecture_keyword, token.architecture_simple_name)
        self.solution = 'Reduce spaces between *architecture* keyword and architecture_simple_name to a single space.'




from vsg import token

from vsg.rules import whitespace_before_token

lTokens = []
lTokens.append(token.simple_variable_assignment.assignment)
lTokens.append(token.conditional_variable_assignment.assignment)
lTokens.append(token.selected_variable_assignment.assignment)


class rule_003(whitespace_before_token):
    '''
    This rule checks for at least a single space before the assignment.

    **Violation**

    .. code-block:: vhdl

         counter:= 0;
         count := counter + 1;

    **Fix**

    .. code-block:: vhdl

         counter := 0;
         count := counter + 1;
    '''
    def __init__(self):
        whitespace_before_token.__init__(self, 'variable_assignment', '003', lTokens)
        self.solution = 'Ensure at least a single space exists before the *:=* keyword.'




from vsg import token

from vsg.rules import single_space_between_token_pairs

lTokens = []
lTokens.append([token.use_clause.keyword, token.use_clause.selected_name])


class rule_006(single_space_between_token_pairs):
    '''
    This rule checks for excessive spaces after the **use** keyword.

    **Violation**

    .. code-block:: vhdl

       library ieee;
         use    ieee.std_logic_1164.all;
         use   ieee.std_logic_unsigned.all;

    **Fix**

    .. code-block:: vhdl

       library ieee;
         use ieee.std_logic_1164.all;
         use ieee.std_logic_unsigned.all;
    '''
    def __init__(self):
        single_space_between_token_pairs.__init__(self, 'library', '006', lTokens)
        self.solution = 'Ensure a single space between the *use* keyword and the selected_name.'




from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment
from vsg.vhdlFile import utils


class multiline_alignment_between_tokens(alignment.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    trigger : parser object type
       object type to apply the case check against
    '''

    def __init__(self, name, identifier, lTokenPairs, bExcludeLastToken=False):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.phase = 4
        self.lTokenPairs = lTokenPairs
        self.bExcludeLastToken = bExcludeLastToken
        self.align_left = 'no'
        self.configuration.append('align_left')
        self.align_paren = 'yes'
        self.configuration.append('align_paren')

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in self.lTokenPairs:
            aToi = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1], bExcludeLastToken=self.bExcludeLastToken)
            lToi = utils.combine_two_token_class_lists(lToi, aToi)
        return lToi

    def analyze(self, oFile):
        lToi = self._get_tokens_of_interest(oFile)

        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)
#            print('='*5 + str(iLine) + '='*70)

            iFirstLine, iFirstLineIndent = _get_first_line_info(iLine, oFile)

            iAssignColumn = oFile.get_column_of_token_index(oToi.get_start_index())
            iColumn = iAssignColumn

            dActualIndent = {}

            dActualIndent[iLine] = iFirstLineIndent
            lParens = []
            dIndex = {}

            bStartsWithParen = _starts_with_paren(lTokens)
            bSkipCommentLine = False

            for iToken, oToken in enumerate(lTokens):

                iLine = utils.increment_line_number(iLine, oToken)

                if isinstance(oToken, parser.blank_line):
                    continue

                if bSkipCommentLine:
                   if not isinstance(oToken, parser.carriage_return):
                       continue

                if isinstance(oToken, parser.carriage_return):
                    iColumn = 0
                    bSkipCommentLine = rules_utils.does_line_start_with_comment(lTokens[iToken + 1:iToken + 3])
                    if bSkipCommentLine:
                        dActualIndent[iLine] = None
                    else:
                        dActualIndent[iLine] = _set_indent(iToken, lTokens)
                        dIndex[iLine] = iToken + 1
                    continue

                iColumn += len(oToken.get_value())

                if isinstance(oToken, parser.close_parenthesis):
                    dParen = {}
                    dParen['type'] = 'close'
                    dParen['line'] = iLine
                    dParen['column'] = iColumn
                    dParen['begin_line'] = utils.does_token_start_line(iToken, lTokens)
                    lParens.append(dParen)

                if isinstance(oToken, parser.open_parenthesis):
                    dParen = {}
                    dParen['type'] = 'open'
                    dParen['line'] = iLine
                    dParen['column'] = iColumn
                    lParens.append(dParen)

            iLastLine = iLine

            if iFirstLine == iLastLine:
                continue

            iFirstTokenLength = len(lTokens[0].get_value())

            if self.align_paren == 'no' and self.align_left == 'yes':
                dExpectedIndent = _analyze_align_paren_no(iFirstLine, iLastLine, lParens, self.indentSize, dActualIndent, bStartsWithParen)
            if self.align_paren == 'yes' and self.align_left == 'no':
                dExpectedIndent = _analyze_align_paren_yes_align_left_no(iFirstLine, iLastLine, lParens, dActualIndent, self.indentSize, bStartsWithParen, iAssignColumn, iFirstTokenLength)
            if self.align_paren == 'yes' and self.align_left == 'yes':
                dExpectedIndent = _analyze_align_paren_yes_align_left_yes(iFirstLine, iLastLine, lParens, dActualIndent, self.indentSize, bStartsWithParen, iAssignColumn)
            if self.align_paren == 'no' and self.align_left == 'no':
                dExpectedIndent = _analyze_align_paren_no_align_left_no(iFirstLine, iLastLine, lParens, dActualIndent, self.indentSize, bStartsWithParen, iAssignColumn)


#            print(f'Actual = {dActualIndent}')
#            print(f'Expect = {dExpectedIndent}')
#            print(f'Index  = {dIndex}')

            for iLine in range(iFirstLine, iLastLine + 1):
                if dActualIndent[iLine] is None:
                    continue
                if indents_match(dActualIndent[iLine], dExpectedIndent[iLine]):
                    continue

                oViolation = build_violation(iLine, oToi, iToken, dExpectedIndent, dIndex, dActualIndent)
                self.add_violation(oViolation)


    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
#        print(dAction)
        if dAction['action'] == 'adjust':
            lTokens[0].set_value(' '*dAction['column'])
        else:
            rules_utils.insert_whitespace(lTokens, 0, dAction['column'])

        oViolation.set_tokens(lTokens)


def build_violation(iLine, oToi, iToken, dExpectedIndent, dIndex, dActualIndent):
    sSolution = 'Adjust indent to column ' + str(dExpectedIndent[iLine])
    dAction = build_action_dictionary(iLine, dActualIndent, dExpectedIndent)
    iToken = dIndex[iLine]
    oViolation = violation.New(iLine, oToi.extract_tokens(iToken, iToken), sSolution)
    oViolation.set_action(dAction)
    return oViolation


def build_action_dictionary(iLine, dActualIndent, dExpectedIndent):
    dAction = {}
    dAction['line'] = iLine
    dAction['column'] = dExpectedIndent[iLine]

    if dActualIndent[iLine] > 0:
        dAction['action'] = 'adjust'
    else:
        dAction['action'] = 'insert'
    return dAction


def line_starts_with_comment(iActual):
    if iActual is None:
        return True
    return False


def indents_match(iActual, iExpected):
    if iActual == iExpected:
        return True
    return False


def calculate_start_column(oFile, oToi):
    iReturn = oFile.get_column_of_token_index(oToi.get_start_index())
    iReturn += len(oToi.get_tokens()[0].get_value())
    iReturn += 1
#    print(f'Start Column = {iReturn}')
    return iReturn


def is_token_before_carriage_return(tToken, lTokens):
    for oToken in lTokens:
        if isinstance(oToken, tToken):
            return True
        if isinstance(oToken, parser.carriage_return):
            return False
    return False


def _set_column_adjustment(iToken, lTokens):
    iReturn = 0
    if isinstance(lTokens[iToken + 1], parser.whitespace):
        if isinstance(lTokens[iToken + 2], parser.close_parenthesis):
            iReturn = -1
    else:
        if isinstance(lTokens[iToken + 1], parser.close_parenthesis):
            iReturn = -1
    return iReturn


def _set_indent(iToken, lTokens):
    iReturn = 0
    if isinstance(lTokens[iToken + 1], parser.whitespace):
        iReturn = len(lTokens[iToken + 1].get_value())
    else:
        iReturn = 0
    return iReturn


def _analyze_align_paren_no(iFirstLine, iLastLine, lParens, iIndentStep, dActualIndent, bStartsWithParen):
#    print('--> _analyze_align_paren_no <-' + '-'*70)
    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]
    if bStartsWithParen:
        iFirstIndent = dActualIndent[iFirstLine]
    else:
        iFirstIndent = dActualIndent[iFirstLine] + iIndentStep

    iIndent = iFirstIndent

#    print(iIndent)
    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <--------------------------')

        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

        if len(lTemp) == 0:
            dExpectedIndent[iLine + 1] = iIndent
            continue

        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
            else:
                iParens -= 1
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] -= iIndentStep

        iIndent = iFirstIndent + iParens * iIndentStep

#        print(f'indent = {iIndent} | iPerens = {iParens}')
        dExpectedIndent[iLine + 1] = iIndent
#        print(dExpectedIndent)

    return dExpectedIndent


def _analyze_align_paren_yes_align_left_no(iFirstLine, iLastLine, lParens, dActualIndent, iIndentStep, bStartsWithParen, iAssignColumn, iFirstTokenLength):
#    print('--> _analyze_align_paren_yes_align_left_no <-' + '-'*70)
    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]

#    if bStartsWithParen:
#       iIndent = dActualIndent[iFirstLine]
#       iColumn = iIndent
#       lColumn = [dActualIndent[iFirstLine]]
#    else:
    iIndent = iAssignColumn + iFirstTokenLength + 1
    iColumn = iIndent
    lColumn = [iIndent]

#    print('*'*80)
#    print(lParens)

    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <--------------------------')
        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

#        print(f'lTemp = {lTemp}')
        iTemp = lColumn[-1]
        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
                if iLine == iFirstLine:
                    iColumn = dTemp['column']
                else:
                    iColumn = dTemp['column'] + (iTemp - dActualIndent[iLine]) + iIndentStep - 1
#                print(f"iColumn = {dTemp['column']} + ({iTemp} - {dActualIndent[iLine]}) + {iIndentStep} - 1 = {iColumn}")
                lColumn.append(iColumn)
                dExpectedIndent[iLine + 1] = iColumn
            else:
                iParens -= 1
                lColumn.pop()
                dExpectedIndent[iLine + 1] = lColumn[-1]
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] = dExpectedIndent[iLine] - iIndentStep

#        print(f'iParens = {iParens}')


        if len(lTemp) == 0:
            if bStartsWithParen:
                dExpectedIndent[iLine + 1] = lColumn[-1]
            else:
                if iParens == 0:
                    dExpectedIndent[iLine + 1] = lColumn[-1]
                else:
                    dExpectedIndent[iLine + 1] = lColumn[-1]

#        print(f'{iLine} | {lColumn} | {dExpectedIndent}')

    return dExpectedIndent


def _analyze_align_paren_yes_align_left_yes(iFirstLine, iLastLine, lParens, dActualIndent, iIndentStep, bStartsWithParen, iAssignColumn):
#    print('--> _analyze_align_paren_yes_align_left_yes <-' + '-'*70)
    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]

    if bStartsWithParen:
       iIndent = dActualIndent[iFirstLine]
       iColumn = iIndent
       lColumn = [dActualIndent[iFirstLine]]
    else:
       iIndent = iAssignColumn + 2 + 1
       iColumn = iIndent
       lColumn = [iIndent]

#    print('*'*80)
#    print(lParens)

    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <--------------------------')
        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

#        print(f'lTemp = {lTemp}')
        iTemp = lColumn[-1]
        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
                if iLine == iFirstLine:
                    iColumn = dTemp['column'] + iIndentStep - 1
                else:
                    iColumn = dTemp['column'] + (iTemp - dActualIndent[iLine]) + iIndentStep - 1
#                print(f"iColumn = {dTemp['column']} + ({iTemp} - {dActualIndent[iLine]}) + {iIndentStep} - 1 = {iColumn}")
                lColumn.append(iColumn)
                dExpectedIndent[iLine + 1] = iColumn
            else:
                iParens -= 1
                lColumn.pop()
                dExpectedIndent[iLine + 1] = lColumn[-1]
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] = dExpectedIndent[iLine] - iIndentStep

#        print(f'iParens = {iParens}')

        if len(lTemp) == 0:
            if bStartsWithParen:
                dExpectedIndent[iLine + 1] = lColumn[-1]
            else:
                if iParens == 0:
                    dExpectedIndent[iLine + 1] = lColumn[-1]
                else:
                    dExpectedIndent[iLine + 1] = lColumn[-1]

        if iLine == iFirstLine:
            dExpectedIndent[iLine + 1] = iParens * iIndentStep + dActualIndent[iFirstLine] + iIndentStep
            lColumn[-1] = iParens * iIndentStep + dActualIndent[iFirstLine] + iIndentStep
            if iParens == 0 and not bStartsWithParen:
                dExpectedIndent[iLine + 1] = iIndentStep + dActualIndent[iFirstLine]
                lColumn[-1] = iIndentStep + dActualIndent[iFirstLine]
            elif iParens == 0 and bStartsWithParen:
                dExpectedIndent[iLine + 1] = dActualIndent[iFirstLine]
                lColumn[-1] = dActualIndent[iFirstLine]

        else:
            if iParens == 1:
                dExpectedIndent[iLine + 1] = iParens * iIndentStep + dActualIndent[iFirstLine]
                lColumn[-1] = iParens * iIndentStep + dActualIndent[iFirstLine]

#        print(f'{iLine} | {lColumn} | {dExpectedIndent}')

    return dExpectedIndent


def _analyze_align_paren_no_align_left_no(iFirstLine, iLastLine, lParens, dActualIndent, iIndentStep, bStartsWithParen, iAssignColumn):
#    print('--> _analyze_align_paren_no <-' + '-'*70)
    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]
#    print(iIndent)
    if bStartsWithParen:
        iFirstIndent = dActualIndent[iFirstLine]
    else:
        iFirstIndent = iAssignColumn + 2 + 1

    iIndent = iFirstIndent

    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <--------------------------')

        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

        if len(lTemp) == 0:
            dExpectedIndent[iLine + 1] = iIndent
            continue

        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
            else:
                iParens -= 1
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] -= iIndentStep

        if iLine == iFirstLine:
            if bStartsWithParen:
                iIndent = iFirstIndent + iParens * iIndentStep
            else:
                if iParens == 0:
                    iIndent = iFirstIndent + iParens * iIndentStep
                else:
                    iIndent = iFirstIndent + iParens * iIndentStep - iIndentStep + iIndentStep
        else:
            if iParens == 0:
                iIndent = iFirstIndent + iParens * iIndentStep
            else:
                iIndent = iFirstIndent + iParens * iIndentStep + iIndentStep

#        print(f'indent = {iIndent} | iPerens = {iParens}')
        dExpectedIndent[iLine + 1] = iIndent

    return dExpectedIndent


def _starts_with_paren(lTokens):
    iToken = utils.find_next_non_whitespace_token(1, lTokens)
    try:
        if isinstance(lTokens[iToken], parser.open_parenthesis):
            return True
    except IndexError:
        if isinstance(lTokens[0], parser.open_parenthesis):
            return True
    return False


def _get_first_line_info(iLine, oFile):
    lTemp = oFile.get_tokens_from_line(iLine)
    iIndent = len(lTemp.get_tokens()[0].get_value())
    return iLine, iIndent





from vsg import parser
from vsg import violation

from vsg.rule_group import whitespace
from vsg.rules import utils


class single_space_before_token(whitespace.Rule):
    '''
    Checks for a single space before a token.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : token type object list
       A list of tokens to check for a single space before.
    '''

    def __init__(self, name, identifier, lTokens):
        whitespace.Rule.__init__(self, name=name, identifier=identifier)
        self.lTokens = lTokens

    def _get_tokens_of_interest(self, oFile):
        lToi = oFile.get_token_and_n_tokens_before_it(self.lTokens, 2)
        lReturn = remove_toi_that_begins_a_line(lToi)
        return lReturn

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if not whitespace_exists_before_token(lTokens):
                oViolation = create_insert_violation(oToi)
                self.add_violation(oViolation)
            elif utils.whitespace_is_larger_than_a_single_character(lTokens):
                oViolation = create_adjust_violation(oToi)
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        sAction = oViolation.get_action()
        if sAction == 'insert':
            utils.insert_whitespace(lTokens, 2)
        elif sAction == 'adjust':
            lTokens[1].set_value(' ')
        oViolation.set_tokens(lTokens)


def whitespace_exists_before_token(lTokens):
    if isinstance(lTokens[1], parser.whitespace):
        return True
    return False


def create_insert_violation(oToi):
    return create_violation(oToi, 'insert')


def create_adjust_violation(oToi):
    return create_violation(oToi, 'adjust')


def create_violation(oToi, sAction):
    lTokens = oToi.get_tokens()
    sSolution = 'Ensure a single space before ' + lTokens[2].get_value()
    oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
    oViolation.set_action(sAction)
    return oViolation


def remove_toi_that_begins_a_line(lToi):
    lReturn = []
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if not token_begins_a_line(lTokens):
            lReturn.append(oToi)
    return lReturn


def token_begins_a_line(lTokens):
    if isinstance(lTokens[0], parser.carriage_return) and isinstance(lTokens[1], parser.whitespace):
        return True
    if isinstance(lTokens[1], parser.carriage_return):
        return True
    return False




from vsg.rules import single_space_before_token_if_on_same_line_as_token as Rule

from vsg.token import alias_declaration as token


class rule_101(Rule):
    '''
    This rule checks for a single space before the **is** keyword if the : is present.

    **Violation**

    .. code-block:: vhdl

       alias alias_designator : subtype_indication     is name;
       alias alias_designator   is name;

    **Fix**

    .. code-block:: vhdl

       alias alias_designator : subtype_indication is name;
       alias alias_designator   is name;
    '''
    def __init__(self):
        Rule.__init__(self, 'alias_declaration', '101', token.is_keyword, token.colon)
        self.solution = 'Ensure a single space before **is** keyword.'




from vsg import parser

from vsg.rules import single_space_between_tokens

from vsg.token import case_statement as token


class rule_003(single_space_between_tokens):
    '''
    This rule checks for a single space before the **is** keyword.

    **Violation**

    .. code-block:: vhdl

       case data    is

    **Fix**

    .. code-block:: vhdl

       case data is
    '''
    def __init__(self):
        single_space_between_tokens.__init__(self, 'case', '003', parser.todo, token.is_keyword)
        self.solution = 'Reduce spaces between the expression and the *is* keyword.'




from vsg import parser
from vsg import token

from vsg.rules import single_space_after_token as Rule

lTokens = []
lTokens.append(token.if_statement.elsif_keyword)


class rule_005(Rule):
    '''
    This rule checks for a single space after the **elsif** keyword.

    **Violation**

    .. code-block:: vhdl

      elsif(c = '1') then

      elsif   (c = '1') then

      elsif    b = '0' then

    **Fix**

    .. code-block:: vhdl

      elsif (c = '1') then

      elsif (c = '1') then

      elsif b = '0' then
    '''
    def __init__(self):
        Rule.__init__(self, 'if', '005', lTokens)
        self.solution = 'Ensure only a single space exists after the *elsif* keyword.'




from vsg.rules import move_token_next_to_another_token

from vsg.token import package_body as token


class rule_001(move_token_next_to_another_token):
    '''
    This rule checks the **is** keyword is on the same line as the **package** keyword.

    **Violation**

    .. code-block:: vhdl

       package body FIFO_PKG
       is

    **Fix**

    .. code-block:: vhdl

       package body FIFO_PKG is
    '''

    def __init__(self):
        move_token_next_to_another_token.__init__(self, 'package_body', '001', token.package_simple_name, token.is_keyword)
        self.solution = 'Ensure *is* keyword is on the same line as the "package" keyword.'




from vsg import token

from vsg.rules import single_space_after_token

lTokens = []
lTokens.append(token.signal_declaration.colon)


class rule_005(single_space_after_token):
    '''
    This rule checks for a single space after the colon.

    **Violation**

    .. code-block:: vhdl

       signal wr_en :    std_logic;
       signal rd_en :std_logic;

    **Fix**

    .. code-block:: vhdl

       signal wr_en : std_logic;
       signal rd_en : std_logic;
    '''
    def __init__(self):
        single_space_after_token.__init__(self, 'signal', '005', lTokens)
        self.solution = 'Ensure only a signal space after the colon.'




from vsg import parser
from vsg import violation

from vsg.rule_group import whitespace
from vsg.rules import utils


class single_space_after_token(whitespace.Rule):
    '''
    Checks for a single space after a token

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : token type object list
       A list of tokens to check for a single space after.
    '''

    def __init__(self, name, identifier, lTokens):
        whitespace.Rule.__init__(self, name=name, identifier=identifier)
        self.lTokens = lTokens

    def _get_tokens_of_interest(self, oFile):
        lToi = oFile.get_token_and_n_tokens_after_it(self.lTokens, 2)
        lToi = utils.remove_toi_if_token_is_at_the_end_of_the_line(lToi)
        return lToi

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if not whitespace_exists_after_token(lTokens):
                oViolation = create_violation(oToi, 'insert')
                self.add_violation(oViolation)
            elif utils.whitespace_is_larger_than_a_single_character(lTokens):
                oViolation = create_violation(oToi, 'adjust')
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        sAction = oViolation.get_action()
        if sAction == 'insert':
            utils.insert_whitespace(lTokens, 1)
        elif sAction == 'adjust':
            lTokens[1].set_value(' ')
        oViolation.set_tokens(lTokens)


def whitespace_exists_after_token(lTokens):
    if isinstance(lTokens[1], parser.whitespace):
        return True
    return False


def create_violation(oToi, sAdjust):
    lTokens = oToi.get_tokens()
    sSolution = 'Ensure a single space after ' + lTokens[0].get_value()
    oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
    oViolation.set_action(sAdjust)
    return oViolation
